from __future__ import annotations
import os.path
from typing import Optional

from git import (
	Commit,
	GitCommandError,
	InvalidGitRepositoryError,
	NoSuchPathError,
	Repo
)

from shulkr.config import get_config


def get_repo():
	global repo

	if not repo:
		config = get_config()

		try:
			repo = Repo(config.repo_path)

		except NoSuchPathError:
			print('Initializing git')
			repo = Repo.init(config.repo_path)

		except InvalidGitRepositoryError:
			print('Initializing git')
			repo = Repo.init(config.repo_path)

	return repo


def get_blob(commit: Optional[Commit], path: str) -> bytes:
	config = get_config()

	if commit is None:
		p = os.path.join(config.repo_path, path)
		with open(p, 'r') as f:
			return f.read()

	parts = path.split('/')
	curr = commit.tree

	while len(parts) > 0:
		name = parts.pop(0)
		curr = curr[name]

	return curr.data_stream.read().decode()


def head_has_versions() -> bool:
	repo = get_repo()

	try:
		# List tags reachable by HEAD
		tags = repo.git.tag('--merged')
		return len(tags) > 0

	except GitCommandError as e:
		if 'fatal: malformed object name HEAD' in str(e):
			return False

		raise e


repo = None



import os
import sys

from shulkr import create_gitignore, create_version
from shulkr.config import get_config
from shulkr.arguments import parse_args
from shulkr.git import head_has_versions
from shulkr.minecraft.source import detect_mappings
from shulkr.minecraft.version import NoSuchVersionError, Version, load_manifest


def main_uncaught() -> None:
	load_manifest()

	args = parse_args(sys.argv[1:])

	config = get_config()

	config.repo_path = os.path.join(
		os.getcwd(),
		args.repo
	)

	if args.mappings is None:
		if head_has_versions():
			# Use mappings from previous version
			config.mappings = detect_mappings()
		else:
			# Use default
			config.mappings = 'yarn'
	else:
		config.mappings = args.mappings

	try:
		versions = Version.patterns(args.version)
	except NoSuchVersionError as e:
		print(e, file=sys.stderr)
		sys.exit(1)
	except ValueError as e:
		print(e, file=sys.stderr)
		sys.exit(2)

	if len(versions) == 0:
		print('No versions selected', file=sys.stderr)
		sys.exit(3)

	gitignore = os.path.join(config.repo_path, '.gitignore')
	if not os.path.isfile(gitignore):
		create_gitignore()

	for version_id in versions:
		create_version(
			version_id,
			args.mappings,
			args.undo_renamed_vars,
			args.message,
			args.tag
		)


def main() -> None:
	try:
		main_uncaught()
	except KeyboardInterrupt:
		print('Aborted!', file=sys.stderr)


if __name__ == '__main__':
	main()



import os.path
from shulkr.config import get_config

from shulkr.git import get_blob, get_repo, head_has_versions
from shulkr.java import JavaAnalyzationError
from shulkr.java import get_renamed_variables, undo_variable_renames
from shulkr.minecraft.source import generate_sources
from shulkr.minecraft.version import Version


def create_gitignore() -> None:
	print('Creating gitignore')

	repo = get_repo()

	gitignore_path = os.path.join(repo.working_tree_dir, '.gitignore')

	with open(gitignore_path, 'w+') as gitignore:
		to_ignore = ['.yarn', '.DecompilerMC']
		gitignore.write('\n'.join(to_ignore) + '\n')

	repo.git.add('.gitignore')
	repo.git.commit('-m', 'add .gitignore')


def undo_renames() -> None:
	repo = get_repo()

	commit1 = repo.commit('HEAD')
	commit2 = None  # working tree
	diff_index = commit1.diff(commit2)

	updated_count = 0

	# For each file changed (in reverse, with indices, so we can remove elements
	# in the for loop)
	for diff in diff_index:
		# Only process modified files (no new, deleted, ... files)
		if diff.change_type != 'M':
			continue

		# Only process Java files; leave everything else unchanged
		if not diff.a_path.endswith('.java'):
			continue

		source = get_blob(commit1, diff.a_path)
		target = get_blob(commit2, diff.b_path)

		try:
			renamed_variables = get_renamed_variables(source, target)
		except JavaAnalyzationError as e:
			raise Exception(f'{e} [{diff.a_path} -> {diff.b_path}]')

		if renamed_variables is not None:
			updated_target = undo_variable_renames(target, renamed_variables)
			p = os.path.join(repo.working_tree_dir, diff.a_path)
			with open(p, 'w') as f:
				f.write(updated_target)

			updated_count += 1

	print(f'+ Updated {updated_count} files')


def commit_version(
	version: Version,
	undo_renamed_vars: bool,
	message_template: str
) -> None:

	repo = get_repo()

	commit_msg = message_template.strip().replace('{}', str(version))
	if undo_renamed_vars and head_has_versions():
		commit_msg += '\n\nRenamed variables reverted'

	if get_config().mappings == 'mojang':
		repo.git.add('client', 'server')
	else:
		repo.git.add('src')

	repo.git.commit(message=commit_msg)


def tag_version(version: Version) -> None:
	repo = get_repo()

	repo.create_tag(str(version))


def create_version(
	version: Version,
	mappings: str,
	undo_renamed_vars: bool,
	message_template: str,
	tag: bool
) -> None:

	# 1. Generate source code for the current version
	print(f'\nGenerating sources for Minecraft {version}')

	mappings = get_config().mappings
	generate_sources(version, mappings)

	# 2. If there are any previous versions, undo the renamed variables
	if undo_renamed_vars and head_has_versions():
		print('Undoing renamed variables')
		undo_renames()

	# 3. Commit the new version to git
	print('Committing to git')
	commit_version(version, undo_renamed_vars, message_template)

	# 4. Tag
	if tag:
		tag_version(version)



from shulkr.arguments import parse_args


def test_parse_args_with_no_arguments_exits_with_return_code_of_2(mocker):
	sys_exit = mocker.patch('sys.exit')
	parse_args([])
	assert sys_exit.called_once_with(2)


def test_parse_args_with_one_version_has_correct_version():
	assert parse_args(['1.18']).version == ['1.18']


def test_parse_args_with_one_version_has_default_repo():
	assert parse_args(['1.18']).repo == '.'


def test_parse_args_with_one_version_has_default_message():
	assert parse_args(['1.18']).message == 'version {}'


def test_parse_args_with_one_version_sets_undo_renamed_vars_to_false():
	assert not parse_args(['1.18']).undo_renamed_vars


def test_parse_args_with_one_version_sets_tag_to_true():
	assert parse_args(['1.18']).tag


def test_parse_args_with_one_version_and_no_tags_sets_tag_to_false():
	assert not parse_args(['--no-tags', '1.18']).tag



import os

import git
import pytest
from shulkr.config import get_config
from shulkr.git import get_repo

from shulkr.minecraft.version import Version, clear_manifest, load_manifest


MANIFEST_DATA = {
	'latest': {
		'release': '1.0.0',
		'snapshot': '1.0.0'
	},
	'versions': [
		{
			'type': 'release',
			'id': '1.0.0'
		},
		{
			'type': 'snapshot',
			'id': 'abcdef'
		}
	]
}


class TestVersions:
	def __init__(self, snapshot: Version, release: Version) -> None:
		self.snapshot = snapshot
		self.release = release


def create_commit(mocker):
	return mocker.create_autospec(git.Commit)


def create_tag(mocker):
	tag = mocker.create_autospec(git.Tag)

	# To match 'versions' fixture, for testing Version.pattern()
	mocker.patch.object(tag, 'name', 'abcdef')

	return tag


def create_repo(mocker):
	repo = mocker.create_autospec(git.Repo)

	# Make the repo directory useful for testing shulkr.minecraft.source
	repo.working_tree_dir = os.path.abspath('foo')

	return repo


@pytest.fixture
def decompiler(mocker):
	"""
	Mock _setup_decompiler() to return a fake git repo with the current
	working_tree_dir
	"""

	def mocked_setup_decompiler(local_dir: str, _remote_url: str) -> git.Repo:
		"""Create a fake decompiler subdirectory (.yarn or .DecompilerMC)"""

		# It will be located directly under the shulkr repo directory
		repo = get_repo()
		decompiler_dir = os.path.join(repo.working_tree_dir, local_dir)

		# Create a fake git repo, and set the working_tree_dir property
		decompiler_repo = mocker.create_autospec(git.Repo)
		decompiler_repo.working_tree_dir = decompiler_dir

		return decompiler_repo

	# Tell the generate_sources functions to use our fake decompiler creator
	mocker.patch(
		'shulkr.minecraft.source._setup_decompiler',
		new=mocked_setup_decompiler
	)


@pytest.fixture
def empty_repo(mocker, decompiler):
	repo = create_repo(mocker)

	orig_git_tag = repo.git.tag

	def new_git_tag(*args):
		if args == ('--merged',):
			return ''

		return orig_git_tag(*args)

	mocker.patch.object(repo.git, 'tag', new=new_git_tag)

	# get_repo() will return this value
	mocker.patch('shulkr.git.repo', repo)

	return repo


@pytest.fixture
def nonempty_repo(mocker, decompiler):
	repo = create_repo(mocker)

	# Add a fake commit
	commit = create_commit(mocker)
	mocker.patch.object(repo, 'iter_commits', return_value=iter([commit]))

	# Add a fake tag for that commit
	tag = create_tag(mocker)

	orig_git_tag = repo.git.tag

	def new_git_tag(*args):
		if args == ('--merged',):
			return tag.name

		return orig_git_tag(*args)

	mocker.patch.object(repo.git, 'tag', new=new_git_tag)

	# get_repo() will return this value
	mocker.patch('shulkr.git.repo', repo)

	return repo


@pytest.fixture
def yarn_mappings():
	config = get_config()

	prev_mappings = config.mappings
	config.mappings = 'yarn'

	yield

	config.mappings = prev_mappings


@pytest.fixture
def mojang_mappings():
	config = get_config()

	prev_mappings = config.mappings
	config.mappings = 'mojang'

	yield

	config.mappings = prev_mappings


@pytest.fixture
def versions():
	load_manifest(MANIFEST_DATA, earliest_supported_version_id='abcdef')

	snapshot = Version.of('abcdef')
	release = Version.of('1.0.0')

	yield TestVersions(snapshot, release)

	clear_manifest()


@pytest.fixture
def root_dir():
	script_dir = os.path.dirname(__file__)
	return os.path.realpath(
		os.path.join(script_dir, '..', '..')
	)



from typing import Dict, List, Tuple, Union

import javalang
from javalang.ast import Node
from javalang.tokenizer import Identifier, tokenize
from javalang.tree import MemberReference, VariableDeclaration


class JavaAnalyzationError(Exception):
	pass


def _ast_nodes_equal(
	node1: Node,
	node2: Node,
	recursive: bool,
	matches: List[Tuple[any, any]],
	_first_level=True
) -> bool:

	# Non-recursive comparison; we're done now that the first node has been
	# processed
	if not recursive and not _first_level:
		return True

	if type(node1) is not type(node2):
		return False

	if type(node1) is list:
		if len(node1) != len(node2):
			return False

		for i in range(len(node1)):
			if not _ast_nodes_equal(node1[i], node2[i], recursive, matches, False):
				return False

		return True

	elif isinstance(node1, Node):
		ignored_attrs = [
			attrs for a, b, attrs in matches if node1 in (a, b) or node2 in (a, b)
		]
		for attr in node1.attrs:
			if attr in ignored_attrs:
				continue

			attr1 = getattr(node1, attr)
			attr2 = getattr(node2, attr)

			if not _ast_nodes_equal(attr1, attr2, recursive, matches, False):
					return False

		return True

	else:
		# if node1 != node2:
			# print('XD', node1, node2, '\n')
		return node1 == node2


def ast_nodes_equal(
	node1: Node,
	node2: Node,
	recursive=True,
	matches: List[Tuple[any, any]] = tuple()
) -> bool:

	return _ast_nodes_equal(node1, node2, recursive, matches)


def ast_paths_equal(
	path1: List[Union[Node, List[Node]]],
	path2: List[Union[Node, List[Node]]],
	matches: List[Tuple[any, any]] = tuple()
) -> bool:
	"""
	Determine if two AST node paths are equal, ignoring other parts of the file

	Args:
		path1 (List[Union[Node, List[Node]]]): [description]
		path2 (List[Union[Node, List[Node]]]): [description]
		matches (List[Tuple[any, any]], optional): List of node pairs to
			consider equal. Defaults to tuple().

	Returns:
		bool: [description]
	"""

	if len(path1) != len(path2):
		return False

	# Skip list entries in the path (because they contain irrelevant parts)
	for i in range(len(path1)):
		if not ast_nodes_equal(path1[i], path2[i], recursive=False, matches=matches):
			return False

	return True


def _filter_ast_node(node: Node, t: type, _path=[]) -> List[Union[Node, str]]:
	if node is None:
		return

	elif type(node) is list:
		for item in node:
			for path, child in _filter_ast_node(item, t, _path=_path):
				yield path, child

	elif isinstance(node, Node):
		if isinstance(node, t):
			yield _path, node

		for attr in node.attrs:
			yield from _filter_ast_node(
				getattr(node, attr),
				t,
				_path=_path + [node, attr]
			)


def filter_ast_node(node: Node, t: type) -> List[Union[Node, str]]:
	return _filter_ast_node(node, t)


def chunk_ast_nodes_by_path(
	nodes: List[Tuple[List, Node]]
) -> List[Tuple[List, List[Node]]]:

	chunks = []
	curr_path = None
	curr_chunk = None
	for path, node in nodes:
		if path != curr_path:
			if curr_path is not None:
				chunks.append((curr_path, curr_chunk))

			curr_path = path
			curr_chunk = []

		curr_chunk.append(node)

	if curr_path is not None:
		chunks.append((curr_path, curr_chunk))

	return chunks


def _have_same_references(
	a_var: str,
	b_var: str,
	a_refs: List[Tuple],
	b_refs: List[Tuple],
	matches: List[Tuple]
) -> bool:

	ar = [
		(path, ref)
		for path, refs in a_refs
		for ref in refs
		if ref.member == a_var
	]

	br = [
		(path, ref)
		for path, refs in b_refs
		for ref in refs
		if ref.member == b_var
	]

	if len(ar) != len(br):
		return False

	for i in range(len(ar)):
		a_ref_path, a_ref_node = ar[i]
		b_ref_path, b_ref_node = br[i]
		ref_match = (a_ref_node, b_ref_node, 'name')

		if not ast_paths_equal(
			a_ref_path,
			b_ref_path,
			matches=matches + [ref_match]
		):
			return False

	return True


def get_renamed_variables(
	source_code: str,
	target_code: str
) -> Dict[str, str]:

	try:
		source_tree = javalang.parse.parse(source_code)
		target_tree = javalang.parse.parse(target_code)
	except javalang.parser.JavaSyntaxError:
		return None

	source_declarations = chunk_ast_nodes_by_path(
		filter_ast_node(source_tree, VariableDeclaration)
	)

	target_declarations = chunk_ast_nodes_by_path(
		filter_ast_node(target_tree, VariableDeclaration)
	)

	target_var_names = [
		declarator.name
		for path, declarations in target_declarations
		for declaration in declarations
		for declarator in declaration.declarators
	]

	source_references = chunk_ast_nodes_by_path(
		filter_ast_node(source_tree, MemberReference)
	)

	target_references = chunk_ast_nodes_by_path(
		filter_ast_node(target_tree, MemberReference)
	)

	renamed_var_names = []
	matches = []

	# Find all the semantically equalivalent variables that have different
	# names. For the variable declarations a and b to be semantically
	# equalivalent:
	#
	# - They must have be in the same semantic section of the file
	# - They must have the same references and assignments

	# Iterate over every pair of chunks, including the pair of the same chunk.
	for i in range(len(source_declarations)):
		source_decn_path, source_decn_nodes = source_declarations[i]
		for j in range(i, len(target_declarations)):
			target_decn_path, target_decn_nodes = target_declarations[j]

			# 1. Make sure paths would be equal if source_node (a) and
			# target_node (b) were the same.

			# 1a. Must be in the same part of the file
			# First get all previously discovered renamed variables
			prev_matches_here = [
				(old, new, attrs)
				for path, v in matches
				for old, new, attrs in v
				if path in (source_decn_path, target_decn_path)
			]

			# Now make sure the paths are equal (ignoring the previously found
			# renamed variables and the declarations themselves)
			if not ast_paths_equal(
				source_decn_path,
				target_decn_path,
				matches=prev_matches_here
			):
				continue

			for k in range(len(source_decn_nodes)):
				source_decn_node = source_decn_nodes[k]
				for m in range(len(source_decn_node.declarators)):
					source_declarator = source_decn_node.declarators[m]
					# 1b. Must have different names
					if source_declarator.name in target_var_names:
						continue

					sr = [
						(path, ref)
						for path, refs in source_references
						for ref in refs
						if ref.member == source_declarator.name
					]

					for n in range(k, len(target_decn_nodes)):
						target_decn_node = target_decn_nodes[n]
						for q in range(len(target_decn_node.declarators)):
							target_declarator = target_decn_node.declarators[q]
							# 1c. All attributes except the name should be equal
							declarator_match = (source_declarator, target_declarator, 'name')
							if not ast_nodes_equal(
								source_declarator,
								target_declarator,
								recursive=True,
								matches=prev_matches_here + [declarator_match]
							):
								continue

							# 1d. Must have the same references
							if not _have_same_references(
								source_declarator.name,
								target_declarator.name,
								source_references,
								target_references,
								prev_matches_here
							):
								continue

							tr = [
								(path, ref)
								for path, refs in target_references
								for ref in refs
								if ref.member == target_declarator.name
							]

							# (sr and tr have the same length, so we can do this safely)
							for i in range(len(sr)):
								source_ref_path, source_ref_node = sr[i]
								target_ref_path, target_ref_node = tr[i]
								ref_match = (source_ref_node, target_ref_node, 'name')

								for path, mappings in matches:
									if path == target_ref_path:
										mappings.append(ref_match)

								else:
									matches.append((target_ref_path, [ref_match]))

							# 1d. (The function cannot return multiple mappings
							# containing the same variable at the same path)
							other_renames_at_path = False
							never_used = True
							for path, mappings in renamed_var_names:
								if path != target_decn_path:
									continue

								for old, new in mappings:
									if old == source_declarator.name or new == target_declarator.name:
										never_used = False
										break

								if not never_used:
									break

								mappings.append((source_declarator.name, target_declarator.name))
								other_renames_at_path = True
								break

							if not never_used:
								continue

							if not other_renames_at_path:
								mappings = [(source_declarator.name, target_declarator.name)]
								renamed_var_names.append((target_decn_path, mappings))

							for path, v in matches:
								if path == target_decn_path:
									for old, new, _ in v:
										if old == source_declarator:
											raise JavaAnalyzationError(f'duplicate declarator in source: {old}')

										if new == target_declarator:
											raise JavaAnalyzationError(f'duplicate declarator in target: {new}')

									v.append(declarator_match)

							else:
								matches.append((target_decn_path, [declarator_match]))

	return renamed_var_names


def undo_variable_renames(code: str, renamed_var_names: List[Tuple]) -> str:
	lines = code.split('\n')
	tokens = tokenize(code)

	for token in tokens:
		if not isinstance(token, Identifier):
			continue

		for path, mappings in renamed_var_names:
			for old, new in mappings:
				if token.value == new:
					line, col = token.position
					line -= 1
					col -= 1
					lines[line] = lines[line][:col] + old + lines[line][col + len(new):]

	return '\n'.join(lines)



from git import Repo


def test_git_repo_initiated(run_yarn):
	# Make sure this is a git repo. If it's not, an error will be thrown.
	Repo(run_yarn.repo_path)


def test_commits_created(run_yarn):
	repo = Repo(run_yarn.repo_path)
	actual = [commit.message.rstrip() for commit in repo.iter_commits()]
	actual.reverse()

	# Calculate expected commit messages from versions
	expected = ['add .gitignore']
	if run_yarn.undo_renamed_vars:
		expected.append(f'version {run_yarn.versions[0]}')
		expected.extend([
			f'version {version}\n\nRenamed variables reverted'
			for version in run_yarn.versions[1:]
		])
	else:
		expected.extend([f'version {version}' for version in run_yarn.versions])

	assert actual == expected


def test_when_running_with_yarn_gitignore_and_src_are_tracked(run_yarn):
	repo = Repo(run_yarn.repo_path)

	# List files and directories that were changed directly under the root
	root_tree = repo.head.commit.tree
	actual = set([obj.name for obj in (root_tree.trees + root_tree.blobs)])
	expected = set(['.gitignore', 'src'])
	assert actual == expected


def test_when_running_with_decompilermc_gitignore_client_and_server_are_tracked(run_mojang):
	repo = Repo(run_mojang.repo_path)

	# List files and directories that were changed directly under the root
	root_tree = repo.head.commit.tree
	actual = set([obj.name for obj in (root_tree.trees + root_tree.blobs)])
	expected = set(['.gitignore', 'client', 'server'])
	assert actual == expected


def test_tags_created(run_yarn):
	repo = Repo(run_yarn.repo_path)
	actual = set([tag.name for tag in repo.tags])
	expected = set(run_yarn.versions)
	assert actual == expected


def test_when_running_with_yarn_working_tree_clean(run_yarn):
	repo = Repo(run_yarn.repo_path)
	assert len(repo.git.status(porcelain=True)) == 0


def test_when_running_with_decompilermc_working_tree_clean(run_mojang):
	repo = Repo(run_mojang.repo_path)
	assert len(repo.git.status(porcelain=True)) == 0



import os
import shutil
import subprocess

from git import Repo

from shulkr.git import get_repo, head_has_versions
from .version import Version


DECOMPILER_MC_REMOTE_URL = 'https://github.com/hube12/DecompilerMC.git'
YARN_REMOTE_URL = 'https://github.com/FabricMC/yarn.git'


def detect_mappings() -> str:
	repo = get_repo()

	subdirectories = set([tree.name for tree in repo.head.commit.tree.trees])
	if subdirectories == set(['client', 'server']):
		# DecompilerMC does not support merged sources
		return 'mojang'

	if subdirectories == set(['src']):
		# Merged sources
		return 'yarn'

	raise Exception('Unable to detect mappings from previous commit')


def _setup_decompiler(local_dir: str, remote_url: str) -> Repo:
	repo = get_repo()

	decompiler_dir = os.path.realpath(
		os.path.join(repo.working_tree_dir, local_dir)
	)

	if os.path.exists(
		os.path.join(decompiler_dir, '.git')
	):
		# Used cached yarn repo
		return Repo(decompiler_dir)
	else:
		# Clone the yarn repo
		print(f'- Cloning {remote_url} into {decompiler_dir}')
		return Repo.clone_from(remote_url, decompiler_dir)


def _generate_sources_with_yarn(version: Version) -> None:
	repo = get_repo()

	decompiler_repo = _setup_decompiler('.yarn', YARN_REMOTE_URL)

	print(f'- Updating mappings to Minecraft {version}')

	# Get latest versions from remote
	decompiler_repo.git.fetch('--prune')

	decompiler_repo.git.reset('--hard', 'HEAD')
	decompiler_repo.git.clean('-fd')

	# Checkout version branch
	decompiler_repo.git.checkout(f'origin/{version}')

	print('- Running decompiler')

	try:
		# Generate source code
		p = subprocess.run(
			['./gradlew', 'decompileCFR'],
			stdout=subprocess.DEVNULL,
			stderr=subprocess.PIPE,
			cwd=decompiler_repo.working_tree_dir
		)
		if p.returncode != 0:
			raise Exception(p.stderr.decode())

		# Output directory
		dest_src_dir = os.path.join(repo.working_tree_dir, 'src')

		# Remove existing top-level destination directory
		if os.path.exists(dest_src_dir):
			shutil.rmtree(dest_src_dir)

		# Move the generated source code to $repo_dir/src
		shutil.move(
			os.path.join(decompiler_repo.working_tree_dir, 'namedSrc'),
			dest_src_dir
		)

	except BaseException as e:
		# Undo src/ deletions
		if head_has_versions():
			repo.git.restore('src')
		else:
			path = os.path.join(repo.working_tree_dir, 'src')
			if os.path.exists(path):
				shutil.rmtree(path)

		raise e


def _generate_sources_with_mojang(version: Version) -> None:
	"""
	Generate sources with DecompilerMC, which uses Mojang's official mappings

	Args:
		version (Version):

	Raises:
		Exception: If DecompilerMC fails
	"""

	repo = get_repo()

	decompiler_repo = _setup_decompiler('.DecompilerMC', DECOMPILER_MC_REMOTE_URL)

	try:
		for env in ('client', 'server'):
			# Generate source code
			p = subprocess.run(
				[
					'python3',
					'main.py',
					'--mcv',
					str(version),
					'-s',
					env,
					'-c',
					'-f',
					'-q'
				],
				stderr=subprocess.PIPE,
				cwd=decompiler_repo.working_tree_dir
			)
			if p.returncode != 0:
				raise Exception(p.stderr.decode())

			# Top-level destination directory ($repo/client or $repo/server)
			dest_dir = os.path.join(repo.working_tree_dir, env)
			dest_src_dir = os.path.join(dest_dir, 'src')

			# Remove existing top-level destination directory
			if os.path.exists(dest_src_dir):
				shutil.rmtree(dest_src_dir)

			# Make top-level destination directory
			if not os.path.exists(dest_dir):
				os.makedirs(dest_dir)

			# Move the generated source code to $dest_dir/src
			shutil.move(
				os.path.join(decompiler_repo.working_tree_dir, 'src', str(version), env),
				dest_src_dir
			)

	except BaseException as e:
		# Undo src/ deletions
		if head_has_versions():
			repo.git.restore('client', 'server')
		else:
			for env in ('client', 'server'):
				path = os.path.join(repo.working_tree_dir, env, 'src')
				if os.path.exists(path):
					shutil.rmtree(path)

		raise e


def generate_sources(version: Version, mappings: str) -> None:
	if mappings == 'mojang':
		_generate_sources_with_mojang(version)

	elif mappings == 'yarn':
		_generate_sources_with_yarn(version)

	else:
		raise ValueError(f"Invalid mapping type '{mappings}'")



import argparse
from typing import List


def parse_args(args: List[str]) -> argparse.Namespace:
	parser = argparse.ArgumentParser(
		prog='shulkr',
		description='Generate multiple versions of the Minecraft source code'
	)

	parser.add_argument(
		'--mappings',
		type=str,
		required=False,
		help='Mappings for deobfuscation, mojang or yarn (default)'
	)

	parser.add_argument(
		'--repo',
		'-p',
		type=str,
		default='.',
		help='Path to the Minecraft repo (defaults to the current working directory)'
	)

	parser.add_argument(
		'--message',
		'-m',
		type=str,
		default='version {}',
		help="Commit message template (defaults to 'version {}')"
	)

	parser.add_argument(
		'--no-tags',
		'-T',
		dest='tag',
		action='store_false',
		help='Do not tag commits'
	)

	parser.add_argument(
		'--undo-renamed-vars',
		'-u',
		dest='undo_renamed_vars',
		action='store_true',
		help=(
			'Revert local variables that were renamed in new versions to their '
			'original names (experimental)'
		)
	)

	parser.add_argument(
		'version',
		nargs='+',
		type=str,
		help='List of versions to decompile'
	)

	parser.set_defaults(tag=True, undo_renamed_vars=False)

	return parser.parse_args(args)



import pytest

from shulkr.minecraft.version import Version


class TestVersion:
	def test_next_of_snapshot_is_the_release(self, versions):
		assert versions.snapshot.next == versions.release

	def test_next_of_release_is_none(self, versions):
		assert versions.release.next is None

	def test_release_does_not_equal_snapshot(self, versions):
		assert versions.release != versions.snapshot

	def test_snapshot_is_less_than_release(self, versions):
		assert versions.snapshot < versions.release

	def test_release_is_greater_than_snapshot(self, versions):
		assert versions.release > versions.snapshot

	def test_snapshot_to_release_returns_list_containing_snapshot_and_release(self, versions):
		assert versions.snapshot.to(versions.release) == [versions.snapshot, versions.release]

	def test_snapshot_to_snapshot_returns_list_containing_snapshot(self, versions):
		assert versions.snapshot.to(versions.snapshot) == [versions.snapshot]

	def test_snapshot_to_none_returns_list_containing_snapshot_and_release(self, versions):
		assert versions.snapshot.to(None) == [versions.snapshot, versions.release]

	def test_snapshot_to_release_excluding_snapshots_returns_list_containing_release(self, versions):
		assert versions.snapshot.to(versions.release, snapshots=False) == [versions.release]

	def test_release_to_snapshot_raises_version_error(self, versions):
		with pytest.raises(ValueError):
			versions.release.to(versions.snapshot)

	def test_pattern_with_no_range_operator_returns_list_containing_version(self, versions):
		assert Version.pattern('1.0.0') == [versions.release]

	def test_pattern_with_snapshot_two_dots_returns_list_containing_release(self, versions):
		assert Version.pattern('abcdef..') == [versions.release]

	def test_pattern_with_snapshot_three_dots_returns_list_containing_snapshot_and_release(self, versions):
		assert Version.pattern('abcdef...') == [versions.snapshot, versions.release]

	def test_pattern_with_release_two_dots_returns_list_containing_release(self, versions):
		assert Version.pattern('1.0.0..') == [versions.release]

	def test_pattern_with_release_two_dots_release_returns_list_containing_release(self, versions):
		assert Version.pattern('1.0.0..1.0.0') == [versions.release]

	def test_pattern_with_snapshot_three_dots_release_returns_list_containing_snapshot_and_release(self, versions):
		assert Version.pattern('abcdef...1.0.0') == [versions.snapshot, versions.release]

	def test_pattern_with_three_dots_on_empty_repo_throws_value_error(self, empty_repo, versions):
		with pytest.raises(ValueError, match='No commits from which to derive current version'):
			Version.pattern('...') == [versions.snapshot, versions.release]

	def test_pattern_with_two_dots_on_repo_with_snapshot_returns_list_containing_release(self, nonempty_repo, versions):
		assert Version.pattern('..') == [versions.release]

	def test_pattern_with_three_dots_on_repo_with_snapshot_returns_list_containing_release(self, nonempty_repo, versions):
		assert Version.pattern('...') == [versions.release]

	def test_patterns_with_empty_list_returns_empty_list(self, versions):
		assert Version.patterns([]) == []

	def test_patterns_with_one_positive_id_returns_corresponding_version(self, versions):
		assert Version.patterns(['1.0.0']) == [versions.release]

	def test_patterns_with_two_positive_identical_ids_returns_one_version(self, versions):
		assert Version.patterns(['1.0.0', '1.0.0']) == [versions.release]

	def test_patterns_with_one_positive_id_and_the_same_negative_id_returns_empty_list(self, versions):
		assert Version.patterns(['1.0.0', '-1.0.0']) == []

	def test_patterns_with_one_positive_id_and_the_same_negative_id_and_the_same_positive_id_returns_one_version(self, versions):
		assert Version.patterns(['1.0.0', '-1.0.0', '1.0.0']) == [versions.release]



import os
import subprocess

from shulkr.minecraft.source import detect_mappings, generate_sources


class GitTree:
	def __init__(self, name: str = None) -> None:
		self.name = name


class SubprocessMock:
	def __init__(self, returncode=0, stderr=None):
		self.returncode = returncode
		self.stderr = stderr


def test_detect_mappings_with_yarn_commit_returns_yarn(mocker, nonempty_repo):
	mocker.patch.object(nonempty_repo.head.commit.tree, 'trees', [GitTree('src')])

	assert detect_mappings() == 'yarn'


def test_detect_mappings_with_mojang_commit_returns_mojang(mocker, nonempty_repo):
	mocker.patch.object(nonempty_repo.head.commit.tree, 'trees', [GitTree('client'), GitTree('server')])

	assert detect_mappings() == 'mojang'


def test_generate_sources_with_yarn_runs_decompiler(mocker, empty_repo, versions):
	subprocess_run = mocker.patch(
		'subprocess.run',
		return_value=SubprocessMock()
	)
	mocker.patch('shutil.rmtree')
	mocker.patch('shutil.move')
	mocker.patch('os.makedirs')

	generate_sources(versions.snapshot, 'yarn')

	decompiler_dir = os.path.join(empty_repo.working_tree_dir, '.yarn')
	subprocess_run.assert_called_once_with(
		['./gradlew', 'decompileCFR'],
		stdout=subprocess.DEVNULL,
		stderr=subprocess.PIPE,
		cwd=decompiler_dir
	)


def test_generate_sources_with_yarn_moves_sources_to_repo(mocker, empty_repo, versions):
	mocker.patch(
		'subprocess.run',
		return_value=SubprocessMock()
	)
	mocker.patch('shutil.rmtree')
	mocker.patch('os.makedirs')
	move = mocker.patch('shutil.move')

	generate_sources(versions.snapshot, 'yarn')

	decompiler_dir = os.path.join(empty_repo.working_tree_dir, '.yarn')
	move.assert_called_once_with(
		os.path.join(decompiler_dir, 'namedSrc'),
		os.path.join(empty_repo.working_tree_dir, 'src')
	)


def test_generate_sources_with_mojang_runs_decompiler(mocker, empty_repo, versions):
	subprocess_run = mocker.patch(
		'subprocess.run',
		return_value=SubprocessMock()
	)
	mocker.patch('shutil.rmtree')
	mocker.patch('shutil.move')
	mocker.patch('os.makedirs')

	generate_sources(versions.snapshot, 'mojang')

	decompiler_dir = os.path.join(empty_repo.working_tree_dir, '.DecompilerMC')
	calls = [
		mocker.call(
			[
				'python3',
				'main.py',
				'--mcv',
				str(versions.snapshot),
				'-s',
				env,
				'-c',
				'-f',
				'-q'
			],
			stderr=subprocess.PIPE,
			cwd=decompiler_dir
		)
		for env in ('client', 'server')
	]
	subprocess_run.assert_has_calls(calls)


def test_generate_sources_with_mojang_moves_sources_to_repo(mocker, empty_repo, versions):
	mocker.patch(
		'subprocess.run',
		return_value=SubprocessMock()
	)
	mocker.patch('shutil.rmtree')
	mocker.patch('os.makedirs')
	move = mocker.patch('shutil.move')

	generate_sources(versions.snapshot, 'mojang')

	decompiler_dir = os.path.join(empty_repo.working_tree_dir, '.DecompilerMC')
	calls = [
		mocker.call(
			os.path.join(decompiler_dir, 'src', str(versions.snapshot), env),
			os.path.join(empty_repo.working_tree_dir, env, 'src')
		)
		for env in ('client', 'server')
	]
	move.assert_has_calls(calls)



import shulkr
from shulkr.minecraft.version import Version


def test_commit_version_with_yarn_mappings_stages_the_src_directory(empty_repo, yarn_mappings):
	# Call commit_version
	version = Version('1.18.1', 0)
	shulkr.commit_version(version, undo_renamed_vars=False, message_template='{}')

	# src needs to have been staged
	empty_repo.git.add.assert_called_once_with('src')


def test_commit_version_with_mojang_mappings_stages_the_src_directory(empty_repo, mojang_mappings):
	# Call commit_version
	version = Version('1.18.1', 0)
	shulkr.commit_version(version, undo_renamed_vars=False, message_template='{}')

	# client and server need to have been staged
	empty_repo.git.add.assert_called_once_with('client', 'server')


def test_commit_version_creates_a_commit(empty_repo):
	# Call commit_version
	version = Version('1.18.1', 0)
	shulkr.commit_version(version, undo_renamed_vars=False, message_template='{}')

	# commit must have been called
	empty_repo.git.commit.assert_called_once()


def test_commit_version_replaces_all_brackets_with_the_version(empty_repo):
	# Call commit_version
	message_template = '{} {}'
	version = Version('1.18.1', 0)
	shulkr.commit_version(version, undo_renamed_vars=False, message_template=message_template)

	# commit must have been called
	expected_message = message_template.replace('{}', str(version))
	empty_repo.git.commit.assert_called_once_with(message=expected_message)


def test_commit_version_with_existing_commits_and_undo_renamed_vars_adds_note_to_commit_message(nonempty_repo):
	# Call commit_version
	version = Version('1.18.1', 0)
	shulkr.commit_version(version, undo_renamed_vars=True, message_template='{}')

	# commit must have been called
	expected_message = f'{version}\n\nRenamed variables reverted'
	nonempty_repo.git.commit.assert_called_once_with(message=expected_message)


def test_create_version_with_undo_renamed_vars_on_repo_with_no_commits_does_not_call_undo_renames(mocker, empty_repo):
	# Mock
	mocker.patch('shulkr.open', create=True)
	mocker.patch('shulkr.generate_sources')
	mocker.patch('shulkr.undo_renames')
	mocker.patch('shulkr.commit_version')

	# Call create_version
	version = Version('1.18.1', 0)
	shulkr.create_version(version, mappings='yarn', undo_renamed_vars=True, message_template=None, tag=False)

	# Assert that undo_renames was not called
	shulkr.undo_renames.assert_not_called()


def test_create_version_with_undo_renamed_vars_on_repo_with_one_commit_calls_undo_renames(mocker, nonempty_repo):
	# Mock
	mocker.patch('shulkr.open', create=True)
	mocker.patch('shulkr.generate_sources')
	mocker.patch('shulkr.undo_renames')
	mocker.patch('shulkr.commit_version')

	# Call create_version
	version = Version('1.18.1', 0)
	shulkr.create_version(version, mappings='yarn', undo_renamed_vars=True, message_template=None, tag=False)

	# Assert that undo_renames was not called
	shulkr.undo_renames.assert_called_once()


def test_create_version_with_tag_false_does_not_call_tag_version(mocker, empty_repo):
	# Mock
	mocker.patch('shulkr.open', create=True)
	mocker.patch('shulkr.generate_sources')
	mocker.patch('shulkr.commit_version')
	mocker.patch('shulkr.tag_version')

	# Call create_version
	version = Version('1.18.1', 0)
	shulkr.create_version(version, mappings='yarn', undo_renamed_vars=False, message_template='{}', tag=False)

	# tag_version must have been called
	shulkr.tag_version.assert_not_called()


def test_create_version_with_tag_true_calls_tag_version(mocker, empty_repo):
	# Mock
	mocker.patch('shulkr.open', create=True)
	mocker.patch('shulkr.generate_sources')
	mocker.patch('shulkr.commit_version')
	mocker.patch('shulkr.tag_version')

	# Call create_version
	version = Version('1.18.1', 0)
	shulkr.create_version(version, mappings='yarn', undo_renamed_vars=False, message_template='{}', tag=True)

	# tag_version must have been called
	shulkr.tag_version.assert_called_once_with(version)
